/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2d;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.font.FontRenderContext;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.File;
import java.io.FileOutputStream;
import java.security.AccessControlException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JPopupMenu;
import javax.swing.filechooser.FileSystemView;

import com.lowagie.text.Document;
import com.lowagie.text.Rectangle;
import com.lowagie.text.pdf.PdfContentByte;
import com.lowagie.text.pdf.PdfWriter;

 
/** <p>This is a basic viewer for the 2D-scenes in this package. A scene node is
 * displayed in the following order: shape, outline, points, annotation,
 * children.</p>
 * 
 * <p>There are two ways to add geometry to the viewer:  Geometry that is 
 * attached to the viewer window itself (like coordinate systems) can be
 * added to the <code>SceneComponent</code> returned by
 * <code>getBackdrop()</code>. The backdrop is drawn before root and does not
 * participate in picking or encompass. All other geometry should be added to
 * the <code>SceneComponent</code> returned by <code>getRoot()</code>.</p>
 * 
 * <p>The viewer's affine transformation (mapping user space to screen space) is
 * always a diagonal matrix followed by a translation. This means that the
 * visible part in user space is always a rectangle aligned with the axes,
 * called the viewport. If one wants to preserve the aspect ratio of the viewer,
 * <code>encompass(Rectangle2D.Double r)</code> provides a convenient
 * method to make the rectangle <code>r</code> in user space just fit into the
 * viewing window (centered both horizontally and vertically). If the property
 * <code>keepingAspectRatio</code> is <code>false</code>,
 * <code>encompass(r)</code> will make the viewport in fact equal to
 * <code>r</code>. <code>keepingAspectRatio</code> also determines whether the
 * standard scale tool scales both dimensions uniformly or separately.</p>
 * 
 * <p>To implement geometry that dynamically depends on the viewport (like
 * cordinate axes) objects can register with the viewer to be notified about
 * the occurance of a <code>ViewportChangeEvent</code>.</p>
 * 
 * <p>Dragging with the mouse in the viewer window results in drag events being
 * sent to the frontmost (i.e last drawn) drag-enabled scene node that was under
 * the mouse cursor when mouse button was pressed. These events contain
 * information about which part
 * of the node is dragged (the shape, the outline or one of the points) and (as
 * a general modifier) whether the shift key is down or not. One can add
 * behaviour to a scene node either by subclassing and overriding the methods
 * of the <code>DragListener</code> interface or (using a particular
 * implementation of <code>SceneComponent</code> like <code>SceneComponent</code>
 * that supports this) adding a <code>DragListener</code> to the scene node's
 * list of drag listeners.</p>
 * 
 * <p>If no scene node is under the mouse cursor when
 * mouse button is pressed, the drag event is sent to the viewer's primary tool
 * (or the secondary tool if the shift key is down). The primary tool and the
 * secondary tool can be freely set. By default, the primary tool translates
 * the viewer's viewport and the secondary tool scales the viewport. By holding
 * the control key down, one can
 * enforce activation of the viewer tools even when some drag-enabled scene
 * node was under the mouse cursor when mouse button was pressed.</p> */

public class Viewer2D extends JComponent implements AppearanceChangeListener {
	private static final long serialVersionUID = 1984L;

	// minimal lineWidth for picking outlines
	private static int MINIMAL_PICK_LINEWIDTH = 6;
	private static Stroke pickStroke = new BasicStroke(MINIMAL_PICK_LINEWIDTH);


	// How many pixels are left at the window boundary when doing encommpass
	private int encompassMarginX = 20;
	private int encompassMarginY = 20;
	private double snapToGridInterval = .1;
	private int snapToGridResolution = 3;
	private boolean snapToGridEnabled = false;
	private long snapX;
	private long snapY;

	private boolean scaleToolEnabled = true;
	private boolean translateToolEnabled = true;
	private boolean menuToolEnabled = true;

	private JPopupMenu menu = new JPopupMenu();
	private double lastPopupLocationX;
	private double lastPopupLocationY;

	private boolean antialias = true;

	// the root of the tree of scene objects
	private SceneComponent root;

	// this contains geometry attached to the viewer window itself (like
	// coordinate systems). It is drawn before root and does not participate
	// in picking or encompass.
	private SceneComponent backdrop;

	// the transformation that maps world coordinates to screen coordinates
	private AffineTransform transform = AffineTransform.getScaleInstance(1,-1);

	// the transformation that maps screen coordinates to world coordinates
	private AffineTransform inverseTransform = AffineTransform.getScaleInstance(1,-1);

	private AffineTransform dragTransform = AffineTransform.getScaleInstance(1,-1);

	// the node that currently recieves drag events
	private SceneComponent activeNode;

	// the node that currently recieves mouse over events
	private SceneComponent mouseOverNode;

	// the currently active tool of this viewer
	private DragListener viewerTool;

	// the tool that reacts to mouse drags with a translation of viewport
	private DragListener translateTool = new DragListener() {
		double lastX, lastY;
		public void dragStart(TransformedMouseEvent e) {
			lastX = e.getX();
			lastY = e.getY();
		}
		public void drag(TransformedMouseEvent e) {
			double x = e.getX();
			double y = e.getY();
			transform.concatenate(
					AffineTransform.getTranslateInstance(x-lastX, y-lastY)
			);
			lastX = x;
			lastY = y;
			setTransform(transform);
			fireViewportChange();
			repaint();
		}
		public void dragEnd(TransformedMouseEvent e) {
		}
	};

	private DragListener scaleTool = new DragListener() {
		double lastX, lastY, viewportWidth, viewportHeight;
		double viewportCenterX, viewportCenterY;
		public void dragStart(TransformedMouseEvent e) {
			lastX = e.getX();
			lastY = e.getY();
			Rectangle2D.Double viewport = getViewport();
			viewportCenterX = viewport.x + .5 * viewport.width;
			viewportCenterY = viewport.y + .5 * viewport.height;
			viewportWidth = viewport.width;
			viewportHeight = viewport.height;
		}
		public void drag(TransformedMouseEvent e) {
			double x = e.getX();
			double y = e.getY();
			double factorX = Math.exp(2*(x-lastX)/viewportWidth);
			double factorY = Math.exp(2*(y-lastY)/viewportHeight);
			if (keepingAspectRatio) {
				double factor = factorX * factorY;
				factorX = factor;
				factorY = factor;
			}
			transform.concatenate(
					new AffineTransform(
							factorX, 0, 0, factorY,
							(1-factorX)*viewportCenterX, (1-factorY)*viewportCenterY
					)
			);
			lastX = x;
			lastY = y;
			setTransform(transform);
			fireViewportChange();
			repaint();
		}
		public void dragEnd(TransformedMouseEvent e) {
		}
	};

	private DragListener menuTool = new DragListener() {

		public void dragStart(TransformedMouseEvent e) {
			if (e.getMouseEvent().getButton()==3 || menu != null) {
				double x = e.getX();
				double y = e.getY();
				if (snapToGridEnabled) {
					x = snapToGridInterval * Math.round(x/snapToGridInterval);
					y = snapToGridInterval * Math.round(y/snapToGridInterval);
				}
				lastPopupLocationX = x;
				lastPopupLocationY = y;
				MouseEvent m = e.getMouseEvent();
				menu.show(m.getComponent(), m.getX(), m.getY());
			}
		}
		public void drag(TransformedMouseEvent e) {}
		public void dragEnd(TransformedMouseEvent e) {}
	};

	// the list of registered ViewportChangeListeners
	private List<ViewportChangeListener> viewportListeners =
		Collections.<ViewportChangeListener>emptyList();

	// the list of registered DragListeners
	private List<DragListener> dragListeners = Collections.<DragListener>emptyList();

	// whether encompass preserves the aspect ratio
	private boolean keepingAspectRatio = true;

	private int indexOfPickedPart;
	private Rectangle2D.Double initialViewPort;

	private int viewerWidth;

	private int viewerHeight;

	private CoordinateGrid grid = new CoordinateGrid();

	private boolean gridEnabled = false;

	private ViewportChangeListener gridListener = new ViewportChangeListener() {
		public void viewportChange(ViewportChangeEvent e) {
			grid.setRectangle(getViewport());
		}	
	};

	private JFileChooser chooser;
	private Action exportPDFAction;


	public Viewer2D() {
		this(null);
	}

	/** <code>encompassPolicy</code> determines when the viewer will adjust its
	 * viewport to encompass the whole scene. If set to <code>MANUAL<code> this
	 * will only happen if encompass is explicitly called. If set to
	 * <code>ENCOMPASS_ON_FIRST_RESIZE</code>, it will happe in addition the
	 * first time the viewer gets resized (which usually happens the first time
	 * it appears on the screen). If set to ENCOMPASS_ON_RESIZE it will happen
	 * on every resize. */

	public Viewer2D(final Rectangle2D.Double r) {
		initialViewPort = r;
		root = new SceneComponent();
		root.addAppearanceChangeListener(this);

		backdrop = new SceneComponent();
		backdrop.addAppearanceChangeListener(this);
		setName("viewer");
		grid.setName("grid");
		backdrop.addChild(grid);
		setGridEnabled(true);
		final JCheckBoxMenuItem gridItem = new JCheckBoxMenuItem("grid enabled");
		gridItem.setSelected(menuToolEnabled);
		gridItem.addActionListener(
				new ActionListener() {
					public void actionPerformed(ActionEvent e) {
						setGridEnabled(gridItem.isSelected());
					}
				}
		);
		
		setBackground(new Color(225,225,225));
		
		menu.setLightWeightPopupEnabled(false);
		menu.add(gridItem);
		exportPDFAction = new AbstractAction("export pdf") {
			private static final long serialVersionUID = 1L;

			public void actionPerformed(ActionEvent arg0) {
				exportPDF();
			}
		};
		menu.add(exportPDFAction);

		// set up handling of mouse and key events

		addMouseListener(new MouseListener() {
			public void mouseClicked(MouseEvent e) {
				requestFocus();
			}
			public void mousePressed(MouseEvent e) {
				pick(e);
			}
			public void mouseReleased(MouseEvent e) {
				final Point2D.Double p = new Point2D.Double();
				p.x = e.getX();
				p.y = e.getY();
				dragTransform.transform(p, p);
				TransformedMouseEvent f = new TransformedMouseEvent(
						this,
						e,
						indexOfPickedPart,
						p.x,
						p.y
				);
				if (activeNode != null) {
					activeNode.dragEnd(f);
					activeNode = null;
				} else {
					if (viewerTool != null) viewerTool.dragEnd(f);
				}
			}
			public void mouseEntered(MouseEvent e) {
			}
			public void mouseExited(MouseEvent e) {
			}
		});

		addMouseMotionListener(new MouseMotionListener() {
			public void mouseMoved(MouseEvent e) {
				pickForMouseOver(e);
			}
			public void mouseDragged(MouseEvent e) {
				final Point2D.Double p = new Point2D.Double();
				p.x = e.getX();
				p.y = e.getY();
				dragTransform.transform(p, p);
				if (activeNode != null) {
					double x = p.x;
					double y = p.y;
					if (snapToGridEnabled) {
						long newSnapX = Math.round(p.x/snapToGridInterval);
						long newSnapY = Math.round(p.y/snapToGridInterval);
						if (newSnapX != snapX || newSnapY != snapY) {
							snapX = newSnapX;
							snapY = newSnapY;
							x = snapToGridInterval * snapX;
							y = snapToGridInterval * snapY;
							activeNode.drag(
									new TransformedMouseEvent(
											this,
											e,
											indexOfPickedPart,
											x,
											y
									)
							);
						}
					} else {

						activeNode.drag(
								new TransformedMouseEvent(
										this,
										e,
										indexOfPickedPart,
										x,
										y
								)
						);
					}
				} else {
					viewerTool.drag(
							new TransformedMouseEvent(
									this,
									e,
									indexOfPickedPart,
									p.x,
									p.y
							)
					);
				}
			}
		});

		addKeyListener(new KeyAdapter() {
			public void keyTyped(KeyEvent e) {
				if (e.getKeyChar() == 'e') {
					encompass(getBounds2D());
				}
			}
		});

		// make sure that the first time we get a non-zero size the viewing
		// transformation is adjusted in such a way that the whole world is
		// visible

		addComponentListener(
				new ComponentAdapter() {

					public void componentResized(ComponentEvent e) {
						Rectangle2D.Double  rect = initialViewPort;
						if (viewerWidth != 0) {
							// normal behaviour later on
							encompass(getViewport(viewerWidth,viewerHeight),0,0);
						} else {
							// first time to show up
							if (rect == null) {
								if (root.getChildCount() != 0) {
									rect = getBounds2D();
								}
								if (rect == null) {
									rect = new Rectangle2D.Double(-1,-1,2,2);
								}
							}
						}
						encompass(rect);
						viewerWidth = getWidth();
						viewerHeight = getHeight();
					}
				}
		);
		// TODO get lastDir from preferences
		try {
			File lastDir = null;
			FileSystemView view = FileSystemView.getFileSystemView();
			chooser = new JFileChooser((lastDir == null || !lastDir.exists()) ? view
					.getHomeDirectory() : lastDir, view);			
		} catch(AccessControlException e) {
			e.printStackTrace();
		}
	}

	public Rectangle2D.Double getInitialViewPort() {
		return initialViewPort;
	}

	public void setInitialViewPort(Rectangle2D.Double initialViewPort) {
		this.initialViewPort = initialViewPort;
	}

	/** Returns the root of the tree of scene nodes. Edit its attributes to
	 * control the default values of scene node properties */

	public SceneComponent getRoot() {
		return root;
	}

	/** Returns the root node of the part of the scene that contains geometry
	 * attached to the viewer window itself (like coordinate systems). The
	 * backdrop is drawn before root and does not participate in picking or
	 * encompass. */

	public SceneComponent getBackdrop() {
		return backdrop;
	}

	/** Returns the DragListener that is scales the viewport. The translate tool itself
	 * does not check the specifics of the mouseEvent of the dragEvent.
	 * */

	public DragListener getTranslateTool() {
		return translateTool;
	}

	/** Returns the boolean that determines whether left mouse drag on the viewer
	 * background is hooked to translation of the viewport.
	 * */

	public boolean getTranslateToolEnabled() {
		return translateToolEnabled;
	}

	/** Determines whether left mouse drag on the viewer background is hooked to
	 * translation of the viewport.
	 * */

	public void setTranslateToolEnabled(boolean b) {
		translateToolEnabled = b;
	}

	/** Returns the DragListener that is scales the viewport. The scale tool itself
	 * does not check the specifics of the mouseEvent of the dragEvent.
	 * */

	public DragListener getScaleTool() {
		return scaleTool;
	}

	/** Returns the boolean that determines whether left mouse drag on the viewer
	 * background is hooked to scaling of the viewport.
	 * */

	public boolean getScaleToolEnabled() {
		return scaleToolEnabled;
	}

	/** Determines whether left mouse drag on the viewer background is hooked to
	 * scaling of the viewport.
	 * */

	public void setScaleToolEnabled(boolean b) {
		scaleToolEnabled = b;
	}



	/** Returns the DragListener that pops up the context menu. The menu tool itself
	 * does not check the specifics of the mouseEvent of the dragEvent.
	 * */

	public DragListener getMenuTool() {
		return menuTool;
	}

	/** Returns the boolean that determines whether left mouse drag on the viewer
	 * background is hooked to scaling of the viewport.
	 * */

	public boolean getMenuToolEnabled() {
		return menuToolEnabled;
	}

	/** Determines whether right click on the viewer background is hooked to
	 * showing the context menu and setting lastPopupLocationX, lastPopupLocationY.
	 * */

	public void setMenuToolEnabled(boolean b) {
		menuToolEnabled = b;
	}

	public boolean isKeepingAspectRatio() {
		return keepingAspectRatio;
	}

	public void setKeepingAspectRatio(boolean b) {
		keepingAspectRatio = b;
		grid.setKeepingAspectRatio(b);
	}

	/** Requires g to be of type <code>java.awt.Graphics2D</code>. */

	protected void paintComponent(Graphics g) {
		int height = getHeight();
		int width = getWidth();

		if (height == 0 || width == 0) {
			return;
		} else {
			Graphics2D g2d = (Graphics2D) (getComponentGraphics(g).create());
			g2d.setColor(getBackground());
			g2d.fillRect(0, 0, width, height);
			g2d.setRenderingHint(
					RenderingHints.KEY_ANTIALIASING,
					antialias ?
							RenderingHints.VALUE_ANTIALIAS_ON :
								RenderingHints.VALUE_ANTIALIAS_OFF
			);
			g2d.setRenderingHint(
					RenderingHints.KEY_INTERPOLATION,
					RenderingHints.VALUE_INTERPOLATION_BILINEAR
			);
			paintSceneComponent(
					backdrop,
					g2d,
					new SceneAttributes(transform,inverseTransform)
			);
			paintSceneComponent(
					root,
					g2d,
					new SceneAttributes(transform,inverseTransform)
			);
		}
	}

	/** Return the visible Rectangle */

	public Rectangle2D.Double getViewport() {
		return getViewport(getWidth(),getHeight());
	}

	private Rectangle2D.Double getViewport(int width, int height) {
		if (width == 0 || height == 0) {
			return new Rectangle2D.Double(-1, -1, 2, 2);
		}
		Point2D.Double lowerRight = new Point2D.Double(width, height);
		try {
			transform.inverseTransform(lowerRight, lowerRight);
		} catch (NoninvertibleTransformException e) {
			System.err.println("transform not invertible");
		}
		Point2D.Double upperLeft = new Point2D.Double(0, 0);
		try {
			transform.inverseTransform(upperLeft, upperLeft);
		} catch (NoninvertibleTransformException e) {
			System.err.println("transform not invertible");
		}
		return new Rectangle2D.Double(upperLeft.x, lowerRight.y, lowerRight.x
				- upperLeft.x, upperLeft.y - lowerRight.y);
	}

	public void encompass(Rectangle2D.Double r) {
		encompass(r,encompassMarginX, encompassMarginY);
	}

	/** Use a translation and a scaling (preserving the aspect ratio) to make
	 * the rectangle <code>r</code> in user space just fit into the viewing
	 * window (centered both horizontally and vertically), respecting the value
	 * of the property <code>encompassMargin</code>. */

	public void encompass(Rectangle2D.Double r, double marginX, double marginY) {
		if (r != null) {
			double rectangleWidth = r.width;
			double rectangleHeight = r.height;
			double rectangleX = r.x;
			double rectangleY = r.y;

			if (rectangleWidth == 0 && rectangleHeight == 0 ) {
				rectangleWidth = 1;
				rectangleHeight =1;
			}
			if (rectangleWidth == 0) {
				rectangleWidth = rectangleHeight;
				rectangleX = rectangleX - .5 * rectangleWidth;
			}
			if (rectangleHeight == 0) {
				rectangleHeight = rectangleWidth;
				rectangleY = rectangleY - .5 * rectangleHeight;
			}

			java.awt.Dimension size = getSize();
			double scaleX = (double) (size.width - 2 * marginX)
			/ rectangleWidth;
			double scaleY = (double) (size.height - 2 * marginY)
			/ rectangleHeight;
			if (keepingAspectRatio) {
				double scale = Math.min(scaleX, scaleY);
				scaleX = scale;
				scaleY = scale;
			}
			double tX = .5*size.width  - scaleX*(rectangleX + .5 * rectangleWidth);
			double tY = .5*size.height + scaleY*(rectangleY + .5 * rectangleHeight);
			transform.setTransform(scaleX, 0, 0, -scaleY, tX, tY);
			// calculate inverse transform
			setTransform(transform);
			fireViewportChange();
			repaint();
		}
	}

	/** Returns the boolean that determines whether antialiasing is turned on. */

	public boolean getAntialias() {
		return antialias;
	}

	/** Sets the boolean that determines whether antialiasing is turned on. */

	public void setAntialias(boolean b) {
		antialias = b;
	}

	/** Returns the integer that determines how many pixels on the left and
	 * right window boundary are left as as a margin when a rectangle (or
	 * the whole scene) is encompassed.
	 * */

	public int getEncompassMarginX() {
		return encompassMarginX;
	}

	/** Sets the integer that determines how many pixels on the left and
	 * right window boundary are left as as a margin when a rectangle (or
	 * the whole scene) is encompassed.
	 * */

	public void setEncompassMarginX(int n) {
		encompassMarginX = n;
	}

	/** Returns the integer that determines how many pixels on the top and
	 * bottom window boundary are left as as a margin when a rectangle (or
	 * the whole scene) is encompassed.
	 * */

	public int getEncompassMarginY() {
		return encompassMarginY;
	}

	/** Sets the integer that determines how many pixels on the top and
	 * bottom window boundary are left as as a margin when a rectangle (or
	 * the whole scene) is encompassed.
	 * */

	public void setEncompassMarginY(int n) {
		encompassMarginY = n;
	}

	void setTransform(AffineTransform t) {
		transform = t;
		try {
			inverseTransform = t.createInverse();
		} catch (NoninvertibleTransformException f) {
			System.err.println("transform not invertible");
		}
	}

	void paintSceneComponent(SceneComponent c, Graphics2D g, SceneAttributes a) {
		if (!c.isVisible()) return;
		a = a.copy();
		a.setupForPaint(c);
		a.draw(this, g, c);
		for (int i=0; i<c.getChildCount(); i++) {
			paintSceneComponent(c.getChild(i), g, a);
		}
	}

	class PickHit {
		SceneComponent node;
		MouseEvent mouseEvent;
		int index;
		double x;
		double y;
		AffineTransform transform;
		AffineTransform inverseTransform;
	}

	private void pickForMouseOver(MouseEvent e) {
		PickHit hit = new PickHit();
		hit.index = -3;
		hit.x = e.getX();
		hit.y = e.getY();
		hit.mouseEvent=e;
		hit.transform = transform;
		hit.inverseTransform = inverseTransform;

		pick(root, hit, new SceneAttributes(transform, inverseTransform));
		dragTransform = hit.inverseTransform;
		indexOfPickedPart = hit.index;

		final Point2D.Double p = new Point2D.Double();
		dragTransform.transform(p, p);
		if (mouseOverNode != hit.node) {
			if (mouseOverNode != null) {
				mouseOverNode.mouseLeave(
						new TransformedMouseEvent(
								this,
								hit.mouseEvent,
								indexOfPickedPart,
								hit.x,
								hit.y
						)
				);
			}
			if (hit.node != null) {
				hit.node.mouseEnter(
						new TransformedMouseEvent(
								this,
								hit.mouseEvent,
								indexOfPickedPart,
								hit.x,
								hit.y
						)
				);
			}
			mouseOverNode = hit.node;
		}
	}

	void pick(MouseEvent e) {
		PickHit hit = new PickHit();
		hit.index = -3;
		hit.x = e.getX();
		hit.y = e.getY();
		hit.mouseEvent=e;
		hit.transform = transform;
		hit.inverseTransform = inverseTransform;

		pick(root, hit, new SceneAttributes(transform, inverseTransform));

		dragTransform = hit.inverseTransform;
		indexOfPickedPart = hit.index;

		final Point2D.Double p = new Point2D.Double();
		p.x = hit.x;
		p.y = hit.y;
		dragTransform.transform(p, p);
		if (hit.node == null || e.isControlDown()) {
			viewerTool = null;
			int button = hit.mouseEvent.getButton();
			if (button == 1 && translateToolEnabled) viewerTool = translateTool;
			if (button == 2 && scaleToolEnabled) viewerTool = scaleTool;
			if (button == 3 && menuToolEnabled) viewerTool = menuTool;
			if (viewerTool != null) viewerTool.dragStart(
					new TransformedMouseEvent(
							this,
							hit.mouseEvent,
							indexOfPickedPart,
							p.x,
							p.y
					)
			);
		} else {
			activeNode = hit.node;
			double x = p.x;
			double y = p.y;
			if (snapToGridEnabled) {
				snapX = Math.round(p.x/snapToGridInterval);
				snapY = Math.round(p.y/snapToGridInterval);
				x = snapToGridInterval * snapX;
				y = snapToGridInterval * snapY;
			}
			activeNode.dragStart(
					new TransformedMouseEvent(
							this,
							hit.mouseEvent,
							indexOfPickedPart,
							x,
							y
					)
			);
		}
	}

	void pick(SceneComponent c, PickHit hit, SceneAttributes b) {
		if (!c.isVisible()) return;
		final SceneAttributes a = b.copy();
		a.setupForPick(c);
		Shape currentShape = c.getShape();
		Shape trfShape;
		if (currentShape != null) {
			if (c.isAreaDragEnabled()) {
				trfShape = a.tr.createTransformedShape(currentShape);
				if (trfShape.contains(hit.x, hit.y)) {
					hit.node = c;
					hit.index = TransformedMouseEvent.AREA;
					hit.transform = a.tr;
					hit.inverseTransform = a.inverseTr;
				}
			}
			if (c.isOutlineDragEnabled()) {
				trfShape = a.tr.createTransformedShape(currentShape);
				Stroke stroke;
				if (
						a.stroke instanceof BasicStroke &&
						((BasicStroke)a.stroke).getLineWidth()<MINIMAL_PICK_LINEWIDTH
				) {
					stroke = pickStroke;
				} else {
					stroke = a.stroke;
				}
				trfShape = stroke.createStrokedShape(trfShape);
				if (trfShape.contains(hit.x, hit.y)) {
					hit.node = c;
					hit.index = TransformedMouseEvent.OUTLINE;
					hit.transform = a.tr;
					hit.inverseTransform = a.inverseTr;
				}
			}
		}
		if (c.isPointDragEnabled()) {
			final List<Point2D.Double> points = c.getPoints();
			final Point2D.Double p = new Point2D.Double();
			int j = 0;
			for (Point2D.Double q : points) {
				p.setLocation(q);
				a.tr.transform(p,p);
				if (a.pointShape.contains(hit.x-p.x, hit.y-p.y)) {
					hit.node = c;
					hit.transform = a.tr;
					hit.inverseTransform = a.inverseTr;
					hit.index = j;
				}
				j++;
			}
		}
		if (c.isAnnotationDragEnabled()) {
			final List<Annotation> annotations = c.getAnnotations();
			final Point2D.Double p = new Point2D.Double();
			final Rectangle2D.Double r = new Rectangle2D.Double();
			final double[] offsets = new double[2];
			FontRenderContext frc = ((Graphics2D)getGraphics()).getFontRenderContext();
			int j=0;
			for (Annotation annotation : annotations) {
				p.setLocation(annotation);
				a.tr.transform(p,p);
				int margin = a.annotationMargin.intValue();

				String text=annotation.getText();
				r.setRect(a.font.getStringBounds(text,frc));
				Annotation.writeOffsets(
						margin,
						r,
						annotation.anchor,
						offsets
				);
				r.setRect(
						p.x + offsets[0],
						p.y + offsets[1] + r.y,
						r.width,
						r.height
				);
				if (r.contains(hit.x, hit.y)) {
					hit.node = c;
					hit.transform = a.tr;
					hit.inverseTransform = a.inverseTr;
					hit.index = j;
				}
				j++;
			}
		}
		if (c.isIconDragEnabled()) {
			final List<Icon2D> icons = c.getIcons();
			final Point2D.Double p = new Point2D.Double();
			final Rectangle2D.Double r = new Rectangle2D.Double();
			final double[] offsets = new double[2];
			int j=0;
			for (Icon2D icon2D : icons) {
				Icon icon = icon2D.getIcon();
				p.x = icon2D.getX();
				p.y = icon2D.getY();
				a.tr.transform(p,p);
				int margin = a.annotationMargin.intValue();
				r.setRect(0,0,icon.getIconWidth(),icon.getIconHeight());
				Annotation.writeOffsets(
						margin,
						r,
						icon2D.getAnchor(),
						offsets
				);
				r.setRect(p.x+offsets[0],p.y+offsets[1],icon.getIconWidth(),icon.getIconHeight());
				if (r.contains(hit.x, hit.y)) {
					hit.node = c;
					hit.transform = a.tr;
					hit.inverseTransform = a.inverseTr;
					hit.index = j;
				}
				j++;
			}
		}
		for (int j = 0; j < c.getChildCount(); j++) {
			pick(c.getChild(j), hit, a);
		}
	}

	void fireViewportChange() {
		ViewportChangeEvent e =
			new ViewportChangeEvent(this,getViewport());
		for (int i = 0; i < viewportListeners.size(); i++) {
			((ViewportChangeListener)viewportListeners.get(i)).viewportChange(e);
		}
	}

	/** Adds <code>l</code> to the list of listeners that are notified when the
	 * viewport changes. */

	public void addViewportChangeListener(ViewportChangeListener l) {
		if (viewportListeners == Collections.EMPTY_LIST) {
			viewportListeners = new ArrayList<ViewportChangeListener>();
		}
		viewportListeners.add(l);
	}

	/** Removes <code>l</code> from the list of listeners that are notified
	 * when the viewport changes. */

	public void removeViewportChangeListener(ViewportChangeListener l) {
		int index = viewportListeners.indexOf(l);
		if (index == -1)
			throw new IllegalStateException(l+"is not a viewportListener of this component");
		viewportListeners.remove(index);
	}

	public void appearanceChange(AppearanceChangeEvent e) {
		repaint();
	}

	/** Return the rectangle in user space enclosing the whole scene. */

	public Rectangle2D.Double getBounds2D() {
		return SceneAttributes.getBounds2D(
				root, 
				new SceneAttributes(new AffineTransform(), new AffineTransform())
		);
	}

	public boolean isGridEnabled() {
		return gridEnabled;
	}

	public void setGridEnabled(boolean b) {
		grid.setVisible(b);
		if (b && !gridEnabled) {
			addViewportChangeListener(gridListener);
		}
		if (!b && gridEnabled) {
			removeViewportChangeListener(gridListener);
		}
		gridEnabled = b;
		repaint();
	}

	public CoordinateGrid getGrid() {
		return grid;
	}

	/** Adds <code>l</code> to the list of listeners that are notified of
	 * drag events on the viewer background. */

	public void addDragListener(DragListener l) {
		if (dragListeners == Collections.EMPTY_LIST) {
			dragListeners = new ArrayList<DragListener>();
		}
		dragListeners.add(l);
	}

	/** Removes <code>l</code> from the list of listeners that are notified of
	 * drag events on the viewer background. */

	public void removeDragListener(DragListener l) {
		int index = dragListeners.indexOf(l);
		if (index == -1)
			throw new IllegalStateException(l+"is not a dragListener of this component");
		dragListeners.remove(index);
	}

	public double getLastPopupLocationX() {
		return lastPopupLocationX;
	}

	public double getLastPopupLocationY() {
		return lastPopupLocationY;
	}

	public JPopupMenu getMenu() {
		return menu;
	}

	public boolean isSnapToGridEnabled() {
		return snapToGridEnabled;
	}

	public void setSnapToGridEnabled(boolean snapToGridEnabled) {
		this.snapToGridEnabled = snapToGridEnabled;
	}

	public int getSnapToGridResolution() {
		return snapToGridResolution;
	}

	public void setSnapToGridResolution(int snapToGridResolution) {
		this.snapToGridResolution = snapToGridResolution;
		int n = snapToGridResolution/3;
		int k = snapToGridResolution%3;
		if (k<0) {
			k+=3;
			n--;
		}
		double interval = Math.pow(10, -n);
		if (k==1) interval /= 2;
		if (k==2) interval /= 5;
		snapToGridInterval = interval;
	}

	public void exportPDF() {
		chooser.showSaveDialog(this);
		File f = chooser.getSelectedFile();
		if (f != null) {
			Document document = new Document();
			try {
				PdfWriter writer = PdfWriter.getInstance(
						document,
						new FileOutputStream(f)
				);
				Rectangle rect = new Rectangle(
						(float) getWidth(),
						(float) getHeight()
				);
				document.setPageSize(rect);
				document.open();
				PdfContentByte cb = writer.getDirectContent();	
				cb.saveState();
				java.awt.Graphics2D g2 = cb.createGraphics(getWidth(), getHeight());
				paint(g2);
				g2.dispose();
				cb.restoreState();
			} catch (Exception e) {
				e.printStackTrace();
			} finally {
			document.close();
		}
	}
}

public Action getExportPDFAction() {
	return exportPDFAction;
}
}

