/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.beans;


import java.awt.event.ActionEvent;
import java.awt.geom.Rectangle2D;
import java.beans.Introspector;
import java.security.AccessControlException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.jtem.beans.InspectorPanel;
import de.jtem.java2d.SceneComponent;
import de.jtem.java2d.Viewer2D;

public class Viewer2DWithInspector extends Viewer2D {
	
	private static final long serialVersionUID = 1L;
	final public static Set<String> excludedPropertiesForViewer = new HashSet<String>();
	static {
		String[] s = new String[] {"UIClassID","alignmentX","alignmentY","autoscrolls",
				"componentCount","debugGraphicsOptions","enabled","focusCycleRoot",
				"focusTraversalPolicyProvider","focusTraversalPolicySet","focusable",
				"font","foreground","height","inheritsPopupMenu","managingFocus",
				"name","opaque","optimizedDrawingEnabled","paintingTile",
				"requestFocusEnabled","toolTipText","validateRoot","doubleBuffered",
				"verifyInputWhenFocusTarget","visible","width","x","y","encompassMargin",
				"keepingAspectRatio","translateToolEnabled","scaleToolEnabled",
				"menuToolEnabled","baselineResizeBehavior","lastPopupLocationX",
				"lastPopupLocationY","paintingForPrint"
		};
		for (int j=0; j<s.length; j++) {
			excludedPropertiesForViewer.add(s[j]);
		}
	}
	
	private JMenu inspectMenu = new JMenu("inspect");
	protected InspectorPanel inspector;
    private JFrame inspectorFrame;
    private HashMap<Object,JMenuItem> objectToMenuItem = new HashMap<Object,JMenuItem>();
    private boolean usingOwnInspector;
    private boolean firstInspectorSpawn = true;


	public Viewer2DWithInspector() {
		this(null);
	}

	public Viewer2DWithInspector(Rectangle2D.Double r) {
		this(r,null);
	}
	
	public Viewer2DWithInspector(Rectangle2D.Double r, InspectorPanel i) {
		super(r);
		try {
		    String[] searchPath = Introspector.getBeanInfoSearchPath();
		    ArrayList<String> path = new ArrayList<String>();
		    path.addAll(Arrays.asList(searchPath));
		    path.add("de.jtem.java2dx.beans");
		    String[] newPath = new String[path.size()];
			Introspector.setBeanInfoSearchPath(path.toArray(newPath));
		} catch (AccessControlException e)	{
			e.printStackTrace();
		}
		usingOwnInspector = i==null;
		if (usingOwnInspector) {
			inspector = new InspectorPanel();
			inspectorFrame = new JFrame();
	    	inspectorFrame.getContentPane().add(inspector);
	    	inspectorFrame.setTitle("inspector");
		} else {
			inspector = i;
		}
		inspector.addChangeListener(
    			new ChangeListener() {
    				public void stateChanged(ChangeEvent event) {
    					if (event.getSource() instanceof SceneComponent) {
    						((SceneComponent)event.getSource()).fireAppearanceChange();
    					}
    				}
    			}
    	);
		getMenu().removeAll();
		addInspectActionToMenu(this,"viewer",excludedPropertiesForViewer);
		addInspectActionToMenu(getGrid(),"grid");
		getMenu().add(inspectMenu);
	}
	
	public void addInspectActionToMenu(
			final Object o,
			final String name
	) {
		addInspectActionToMenu(o,name,null);
	}

	public void addInspectActionToMenu(
			final Object o,
			final String name,
			final Set<String> excludedProperties

	) {
		JMenuItem item = inspectMenu.add(new AbstractAction(name) {
			final static long serialVersionUID = 1984;
			public void actionPerformed(ActionEvent e) {
				inspect(o,excludedProperties);
			}
		});
		objectToMenuItem.put(o,item);
	}
	
	public void removeInspectActionFromMenu(Object o) {
		inspectMenu.remove((JMenuItem)objectToMenuItem.get(o));
	}
	
    public void inspect(Object obj) {
    	inspect(obj, null);
    }
    
    public void inspect(Object obj, Set<String> excludedPropertyNames) {
    	inspector.setObject(obj, excludedPropertyNames);
    	if (usingOwnInspector) {
    		if (obj != null) {
    			if (firstInspectorSpawn) {
    				inspectorFrame.setLocationRelativeTo(this);
    				firstInspectorSpawn = false;
    			}
    			inspectorFrame.validate();
    			inspectorFrame.pack();
    			inspectorFrame.setVisible(true);
    		} else {
    			inspectorFrame.setVisible(false);
    		}
    	} 
    }
    
	public void inspectViewer() {
		inspect(this,excludedPropertiesForViewer);
	}
	
	public void inspectGrid() {
		inspect(getGrid());
	}
	
	public Action getInspectAction(final Object o, final String name) {
		return getInspectAction(o,name,Collections.<String>emptySet());
	}
	
    public Action getInspectAction(final Object o, final String name, final Set<String> excludedPropertyNames) {
    	return new AbstractAction(name) {
			private static final long serialVersionUID = 4067181036657963497L;

			public void actionPerformed(ActionEvent e) {
				inspect(o,excludedPropertyNames);
			}
    	};
    }

	public JMenu getInspectMenu() {
		return inspectMenu;
	}
	
	public InspectorPanel getInspectorPanel() {
		return inspector;
	}
}
