/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.modelling;


import java.awt.event.ActionEvent;
import java.awt.geom.GeneralPath;
import java.io.IOException;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.jtem.java2d.DragListener;
import de.jtem.java2d.SceneComponent;
import de.jtem.java2dx.Point2DDouble;
import de.jtem.java2dx.Point2DList;
import de.jtem.java2dx.Polygon2D;
import de.jtem.modelling.ActionTree;

import java.awt.geom.Point2D;

public class DraggablePolygon2D extends ModellingTool2D {
	private static final long serialVersionUID = 7446083249668583237L;
	private  Polygon2D polygon;
	private DraggablePoint2DList controlPoints;
	private boolean drawingPoints = true;
	private boolean storedClosed;
	
	public DraggablePolygon2D() {
		this(3,DraggablePoint2DList.CIRCULAR,0,0,1,true,true);
	}
	
	public DraggablePolygon2D(
			int defaultPointCount,
			int defaultArrangement,
			double defaultTranslationX,
			double defaultTranslationY,
			double defaultRadius,
			boolean closed,
			boolean drawingPoints
	) { super();
		controlPoints= new DraggablePoint2DList(
				defaultPointCount,
				defaultArrangement,
				defaultTranslationX,
				defaultTranslationY,
				defaultRadius
		);
		this.drawingPoints = drawingPoints;
		polygon = new Polygon2D(new Point2DList(),closed);
		setUp();
		polygon.assign(controlPoints.getModel());
		updateViewFromModel();
	}
	
	/**
	 * Does not copy p.
	 * @param p
	 * @param defaultArrangementType
	 * @param defaultTranslationX
	 * @param defaultTranslationY
	 * @param defaultRadius
	 */
	public DraggablePolygon2D(
			Polygon2D p,
			int defaultArrangement,
			double defaultTranslationX,
			double defaultTranslationY,
			double defaultRadius,
			boolean drawingPoints
	) {
		this.drawingPoints = drawingPoints;
		controlPoints = new DraggablePoint2DList(
				p,
				defaultArrangement,
				defaultTranslationX,
				defaultTranslationY,
				defaultRadius
		);
		
		polygon = new Polygon2D(new Point2DList(),p.isClosed());
		setUp();
		updateModelFromController();
		updateViewFromModel();
	}
	
	public DraggablePolygon2D(DraggablePolygon2D p) {
		super(p);
		drawingPoints = p.isDrawingPoints();
		polygon = new Polygon2D(new Point2DList(),p.polygon.isClosed());
		controlPoints = new DraggablePoint2DList(p.getControlPoints());
		setUp();
		polygon.assign(controlPoints.getModel());
		updateViewFromModel();
	}
	
	public DraggablePoint2DList getControlPoints() {
		return controlPoints;
	}

	public Polygon2D getModel() {
		return polygon;
	}

	public SceneComponent getControlScene() {
		return controlPoints.getViewScene();
	}

	public void translate(double x, double y) {
		controlPoints.translate(x,y);
		updateModelFromController();
		fireModelChange();
		updateViewFromModel();
	}

	public void reset() {
		controlPoints.reset();
	}
	
	public int getDefaultArrangement() {
		return controlPoints.getDefaultArrangement();
	}

	public int getDefaultPointCount() {
		return controlPoints.getDefaultPointCount();
	}

	public double getDefaultRadius() {
		return controlPoints.getDefaultRadius();
	}

	public double getDefaultTranslationX() {
		return controlPoints.getDefaultTranslationX();
	}

	public double getDefaultTranslationY() {
		return controlPoints.getDefaultTranslationY();
	}

	public void setDefaultArrangement(int defaultArrangement) {
		controlPoints.setDefaultArrangement(defaultArrangement);
	}

	public void setDefaultPointCount(int defaultPointCount) {
		controlPoints.setDefaultPointCount(defaultPointCount);
	}

	public void setDefaultRadius(double defaultRadius) {
		controlPoints.setDefaultRadius(defaultRadius);
	}

	public void setDefaultTranslationX(double defaultTranslationX) {
		controlPoints.setDefaultTranslationX(defaultTranslationX);
	}

	public void setDefaultTranslationY(double defaultTranslationY) {
		controlPoints.setDefaultTranslationY(defaultTranslationY);
	}
	
	protected void updateViewFromModel() {
		int n = polygon.size();
		if (n>0) {
			if (drawingPoints) {
				List<Point2D.Double> p = getViewScene().getPoints();
				p.clear();
				p.addAll(polygon);
			}
			GeneralPath path = new GeneralPath();
			Point2DDouble p = polygon.get(0);
			path.moveTo((float)p.x,(float)p.y);
			for (int j=1; j<n; j++) {
				p = polygon.get(j);
				path.lineTo((float)p.x,(float)p.y);
			}
			if (polygon.isClosed()) path.closePath();
			getViewScene().setShape(path);
		}
	}
	
	protected void updateModelFromController() {
		polygon.assign(controlPoints.getModel());
	}
	
	protected void updateFromModel() {
		controlPoints.getModel().assign(polygon);
		updateViewFromModel();
	}
	
	private void setUp() {
		addActionTree(new ActionTree(
				new AbstractAction("reset") {
					private static final long serialVersionUID = 1400384559035275571L;
					
					public void actionPerformed(ActionEvent e) {
						reset();
					}
				}));
		storedClosed = getModel().isClosed();
		
		controlPoints.setMenuDragListener(getMenuDragListener());
		controlPoints.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				updateModelFromController();
				fireModelChange();
				updateViewFromModel();
			}
		});
		SceneComponent viewScene = getViewScene();
		viewScene.setAreaDragEnabled(true);
		viewScene.setOutlineDragEnabled(true);
		viewScene.addDragListener(controlPoints.getTransformationTool());
	}

	public boolean isDrawingPoints() {
		return drawingPoints;
	}

	public void setDrawingPoints(boolean b) {
		if (!b && drawingPoints) {
			getViewScene().getPoints().clear();
		}
		drawingPoints = b;
		updateViewFromModel();
	}

	public void storeModelState() {
		controlPoints.storeModelState();
		storedClosed = getModel().isClosed();
	}

	public void recallModelState() {
		controlPoints.recallModelState();
		getModel().setClosed(storedClosed);
		updateModelFromController();
		fireModelChange();
		updateViewFromModel();
	}
	
	public ActionTree getActionTree() {
		return new ActionTree(
				new AbstractAction("reset") {
					static final long serialVersionUID = 6076889467437920180L;

					public void actionPerformed(ActionEvent e) {
						reset();
					}
				}
		);
	}
	
	public DraggablePolygon2D clone() {
		return new DraggablePolygon2D(this);
	}
	
	public void setMenuDragListener(DragListener l) {
		controlPoints.setMenuDragListener(l);
		super.setMenuDragListener(l);
	}
	
	private void readObject(java.io.ObjectInputStream stream)
	throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		setUp();
	}
}
