/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.modelling;


import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Paint;
import java.awt.Stroke;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Rectangle2D;
import java.util.Collections;
import java.util.HashSet;

import javax.swing.JFrame;

import de.jtem.beans.ColorEditor;
import de.jtem.beans.EditorManager;
import de.jtem.beans.FontEditor;
import de.jtem.beans.StrokeEditor;
import de.jtem.java2d.Annotation;
import de.jtem.java2d.SceneComponent;
import de.jtem.java2dx.Line2DDouble;
import de.jtem.java2dx.beans.Viewer2DWithInspector;
import de.jtem.modelling.ActionTree;
import de.jtem.modelling.Modeller;

public class GraphicsModeller2D extends SimpleModeller2D {

	private static final long serialVersionUID = 933354580637372378L;
	
	public GraphicsModeller2D() {
		this(new Viewer2DWithInspector(new Rectangle2D.Double(-3,-3,6,6)));
	}
	
    public GraphicsModeller2D(Viewer2DWithInspector viewer) {
    	super(viewer);
    	EditorManager.registerEditor(Paint.class, ColorEditor.WithNullOption.class);
    	EditorManager.registerEditor(Stroke.class, StrokeEditor.WithNullOption.class);
    	EditorManager.registerEditor(Font.class, FontEditor.WithNullOption.class);
    	getModeller().setCopyEnabled(true);
    	
    	// excluded properties for inspection
    	HashSet<String> excludedPropertiesForPointScene = new HashSet<String>();
    	excludedPropertiesForPointScene.addAll(Modeller2DUtility.getExpertProperties());
    	excludedPropertiesForPointScene.addAll(Modeller2DUtility.getAreaProperties());
    	excludedPropertiesForPointScene.addAll(Modeller2DUtility.getOutlineProperties());
    	excludedPropertiesForPointScene.addAll(Modeller2DUtility.getIconAndAnnotationProperties());
    	excludedPropertiesForPointScene.addAll(Modeller2DUtility.getAnnotationProperties());
   
    	HashSet<String> excludedPropertiesForPolygonScene = new HashSet<String>();
    	excludedPropertiesForPolygonScene.addAll(Modeller2DUtility.getExpertProperties());
    	excludedPropertiesForPolygonScene.addAll(Modeller2DUtility.getAnnotationProperties());
    	excludedPropertiesForPolygonScene.addAll(Modeller2DUtility.getIconAndAnnotationProperties());
    	
    	HashSet<String> excludedPropertiesForLineScene = new HashSet<String>();
    	excludedPropertiesForLineScene.addAll(excludedPropertiesForPolygonScene);
    	excludedPropertiesForLineScene.addAll(Modeller2DUtility.getAreaProperties());
     	
    	HashSet<String> excludedPropertiesForAnnotationScene = new HashSet<String>();
    	excludedPropertiesForAnnotationScene.addAll(Modeller2DUtility.getExpertProperties());
    	excludedPropertiesForAnnotationScene.addAll(Modeller2DUtility.getAreaProperties());
    	excludedPropertiesForAnnotationScene.addAll(Modeller2DUtility.getOutlineProperties());
    	excludedPropertiesForAnnotationScene.addAll(Modeller2DUtility.getPointProperties());
    	
    	HashSet<String> excludedPropertiesForIconScene = new HashSet<String>();
    	excludedPropertiesForIconScene.addAll(excludedPropertiesForAnnotationScene);
    	excludedPropertiesForIconScene.addAll(Modeller2DUtility.getAnnotationProperties());
    	
    	// control scene
    	SceneComponent controlScene = new SceneComponent();
    	controlScene.setName("control scene");
        controlScene.setPointShape(new Ellipse2D.Double(-8,-8,16,16));
        controlScene.setPointPaint(new Color(255,255,0,80));
        controlScene.setOutlinePaint(Color.red);
        
    	// point template
        DraggablePoint2D pointTemplate = new DraggablePoint2D();
        pointTemplate.setName("point");
        pointTemplate.setModelType("Point2D");
        SceneComponent pointScene = new SceneComponent();
        pointScene.setPointPaint(Color.green);
        pointScene.setPointShape(new Ellipse2D.Double(-3,-3,6,6));
    	getModeller().addTemplate(
    			pointTemplate,
    			pointScene
    	);
    	getModeller().setExcludedProperties(
    			pointScene,
    			excludedPropertiesForPointScene
    	);
        
        // line template
        DraggableLine2D line2DTemplate = new DraggableLine2D(
        		new Line2DDouble(-1,0,1,0)
        );
        line2DTemplate.setName("line");
        line2DTemplate.setModelType("Line2D");
        SceneComponent line2DScene = new SceneComponent();
        line2DScene.setStroke(new BasicStroke(2));
        getModeller().addTemplate(
        		line2DTemplate,
        		line2DScene
    	);
        getModeller().setExcludedProperties(
        		line2DScene,
        		excludedPropertiesForLineScene
    	);
        
        // polygon template
        DraggablePolygon2D polygonTemplate = new DraggablePolygon2D();
        polygonTemplate.setName("polygon");
        polygonTemplate.setModelType("Polygon2D");
        SceneComponent polygonScene = new SceneComponent();
        polygonScene.setStroke(new BasicStroke(3));
        polygonScene.setPaint(new Color(128,0,80,30));
        polygonScene.setOutlinePaint(Color.red);
        polygonScene.setPointPaint(Color.yellow);
        getModeller().addTemplate(
    			polygonTemplate,
    			polygonScene
    	);
        getModeller().setExcludedProperties(
        		polygonScene,
        		excludedPropertiesForPolygonScene
    	);
        getModeller().setExcludedProperties(polygonTemplate, Collections.<String>singleton("defaultPointCount"));
        
        //subdivided polygon template
        SubdividedPolygon2D subdividedPolygonTemplate = new SubdividedPolygon2D();
        subdividedPolygonTemplate.setDefaultPointCount(6);
        subdividedPolygonTemplate.reset();
        subdividedPolygonTemplate.setName("subdivided polygon");
        subdividedPolygonTemplate.setModelType("Polygon2D");
        SceneComponent subdividedPolygonScene = new SceneComponent(polygonScene);
        getModeller().addTemplate(
    			subdividedPolygonTemplate,
    			subdividedPolygonScene
    	);
        getModeller().setExcludedProperties(
        		polygonScene,
        		excludedPropertiesForPolygonScene
    	);
    	
      	// annotation template
        DraggableAnnotation annotationTemplate =
        	new DraggableAnnotation(
        			new Annotation("text",0,0,Annotation.CENTER)
        	);
        annotationTemplate.setName("annotation");
        annotationTemplate.setModelType("Annotation");
        SceneComponent annotationScene = new SceneComponent();
        annotationScene.setFont(new Font("Lucida Sans Unicode", 0, 14));
        getModeller().addTemplate(
    			annotationTemplate,
    			annotationScene
    	);
        getModeller().setExcludedProperties(
        		annotationScene,
        		excludedPropertiesForAnnotationScene
    	);
    	
       	// latex formula template
    	try {
        DraggableLatexIcon latexFormulaTemplate =
        	new DraggableLatexIcon(
        			new LatexIcon("$\\int_a^b f(x) dx$",20),
        			0,
        			0,
        			Annotation.CENTER
        	);
        latexFormulaTemplate.setName("latex icon");
        latexFormulaTemplate.setModelType("LatexIcon");
        SceneComponent latexFormulaScene = new SceneComponent();
        getModeller().addTemplate(
    			latexFormulaTemplate,
    			latexFormulaScene
    	);
        getModeller().setExcludedProperties(
        		latexFormulaScene,
        		excludedPropertiesForIconScene
    	);
    	} catch (Exception e){
    		System.out.println("latex is not properly installed.");
    		e.printStackTrace();
    	}
    	setActionTree(getModeller().getViewerMenu());
    }
	
    public static void main(String[] args) {
		JFrame frame = new JFrame("graphics modeller 2D");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setSize(600,600);
		GraphicsModeller2D graphicsModeller = new GraphicsModeller2D();
		Modeller modeller = graphicsModeller.getModeller();
		ActionTree viewerMenu = modeller.getViewerMenu();
		viewerMenu.add(modeller.getFileMenu());
		viewerMenu.add(modeller.getEditMenu());
		viewerMenu.add(modeller.getInspectMenu());
		graphicsModeller.setActionTree(viewerMenu);
		frame.getContentPane().add(graphicsModeller.getViewer());
		frame.setVisible(true);
	}
}
