/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.modelling;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;

import javax.swing.Icon;
import javax.swing.ImageIcon;


public class LatexIcon implements Serializable {
	
	private static final long serialVersionUID = -402576996760609889L;
	private final static String documentStart = 
		"\\documentclass[a4paper, 10pt]{article}\n" +
		"\\usepackage[latin1]{inputenc}\n" +
		"\\usepackage{amssymb}\n" +
		"\\usepackage{amsfonts}\n" +
		"\\pagestyle{empty}\n\n" +
		"\\author{jtem}\n" +
		"\\title{Formel}\n\n" +
		"\\begin{document}\n";
	private String latexSource = "\\LaTeX";
	private transient String log = "";
	private transient Process p = null;
	private transient BufferedImage image;
	private static class PreviewIcon implements Icon {
		LatexIcon latexIcon;
		
		PreviewIcon() {};
		
		PreviewIcon(LatexIcon l) {
			latexIcon = l;
		}

		public void paintIcon(Component c, Graphics g, int x, int y) {
			g.drawImage(
					latexIcon.getImage(),
					x,
					y,
					getIconWidth(),
					getIconHeight(),
					null
			);
		}

		public int getIconWidth() {
			return (int)(latexIcon.getFontSize()*(float)7.227/(float)600*latexIcon.getImage().getWidth());
		}

		public int getIconHeight() {
			return (int)(latexIcon.getFontSize()*(float)7.227/(float)600*latexIcon.getImage().getHeight());
		}
	};
	private PreviewIcon icon = new PreviewIcon(this);
	private int fontSize = 20;

	/**
	 * Creates a <code>LatexIcon</code> with an empty text string and an icon that
	 * is <code>null</code>. This constructor does not call latex or dvipng.
	 */
	public LatexIcon() {}
	
	/**
	 * Creates a <code>LatexIcon</code> 
	 * @param formula
	 * @param dpi
	 */
	public LatexIcon(String latexSource, int fontSize) {
		this.fontSize = fontSize;
		setLatexSource(latexSource);
	}
	
	public LatexIcon(LatexIcon f) {
		fontSize = f.getFontSize();
		setLatexSource(f.getLatexSource());
	}
	
	public BufferedImage getImage() {
		return image;
	}

	public String getLog() {
		return log;
	}
	
	public Icon getIcon() {
		return icon;
	}
	
	public String getLatexSource() {
		return latexSource;
	}
	
	public void setLatexSource(String f) {
		latexSource = f;
		compile();
	}
	
	private void compile() {
		// create a temporary tex file
		File texFile = null;
		try {
			texFile = File.createTempFile("DraggableLatexFormula", ".tex");
			
			// find out the name and parent directory
			String texFileName = texFile.getAbsoluteFile().getName();
			String fileNameWithoutExtension =
				texFileName.substring(0, texFileName.lastIndexOf("."));
			File tmpDir = texFile.getParentFile();
			
			// write tex file
			BufferedWriter out = new BufferedWriter(new FileWriter(texFile));
			out.write(documentStart+latexSource+"\n\\end{document}");
			out.flush();
			out.close();
			
			// execute latex
			String cmd [] =  new String [] {"latex", "-interaction=nonstopmode", texFile.getName()};
			boolean ok = false;
			try {
				p = Runtime.getRuntime().exec(cmd, null, tmpDir);
				ok = true;
				
			} catch(IOException e) {
				System.out.println("failure executing latex");
			}
			if (!ok) return;
			InputStream tex = p.getInputStream();
			int c;
			log = "";
			while( (c = tex.read()) >= 0 ) {
				log += (char)c;
			}
			try {
				p.waitFor();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			File auxFile = new File(tmpDir, fileNameWithoutExtension+".aux");
			if (auxFile.delete());
			File logFile = new File(tmpDir, fileNameWithoutExtension+".log");
			logFile.delete();
			texFile.delete();
			
			// execute dvipng
			dvipng(tmpDir, fileNameWithoutExtension);
			// for some reason dvipng sometimes produces 1x1 images on the first run
			File pngFile = new File(tmpDir, fileNameWithoutExtension+"1.png");
			if (pngFile.exists() && pngFile.length()<100) {
				dvipng(tmpDir, fileNameWithoutExtension);
			}
			File dviFile = new File(tmpDir, fileNameWithoutExtension+".dvi");
			dviFile.delete();
			
			// read in the image
	    	ImageIcon fullScale = new ImageIcon(pngFile.toURL());
	    	pngFile.delete();
	    	image = new BufferedImage(
	    			fullScale.getIconWidth(),
	    			fullScale.getIconHeight(),
	    			BufferedImage.TYPE_INT_ARGB
	    	);
	    	Graphics2D g = image.createGraphics();
	    	g.drawImage(
	    			fullScale.getImage(),
	    			0,
	    			0,
	    			image.getWidth(),
	    			image.getHeight(),
	    			null
	    	);
	    	g.dispose();
		}
		catch(java.io.IOException e) {
			e.printStackTrace();
			return;
		}
	}
	
	private void dvipng(File tmpDir, String fileNameWithoutExtension) {
    	String[] cmd = new String [] {
    			"dvipng",
    			"-D", "600",
    			"-bg", "Transparent",
    			"-T", "tight", 
    			fileNameWithoutExtension
    	};
    	try {
    		final Process q = Runtime.getRuntime().exec(cmd, null, tmpDir);
    		InputStream s = q.getInputStream();
    		int c;
    		while( (c = s.read()) >= 0 ) {
    			log += (char)c;
    		}
    		
    		try {
    			q.waitFor();
    		} catch (InterruptedException e) {
    			e.printStackTrace();
    		}
    	} catch(IOException e) {
    		e.printStackTrace();
    	}
	}
	
	public int getFontSize() {
		return fontSize;
	}

	public void setFontSize(int fontSize) {
		this.fontSize = fontSize;
		compile();
	}
	
	private void readObject(java.io.ObjectInputStream stream)
	throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		compile();
	}
}
