/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.modelling;

import java.awt.Font;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.util.List;
import java.util.NoSuchElementException;

import com.thoughtworks.xstream.converters.ConversionException;
import com.thoughtworks.xstream.converters.Converter;
import com.thoughtworks.xstream.converters.MarshallingContext;
import com.thoughtworks.xstream.converters.UnmarshallingContext;
import com.thoughtworks.xstream.io.HierarchicalStreamReader;
import com.thoughtworks.xstream.io.HierarchicalStreamWriter;
import com.thoughtworks.xstream.mapper.Mapper;

import de.jtem.java2d.SceneComponent;

public class SceneComponentConverter implements Converter {
	
	Mapper mapper;
	
	
	HierarchicalStreamWriter writer;
	MarshallingContext marshallingContext;
	
	HierarchicalStreamReader reader;
	UnmarshallingContext context;
	
	public SceneComponentConverter(Mapper mapper) {
		this.mapper = mapper;
	}
	
	public boolean canConvert(Class type) {
		return type == SceneComponent.class;
	}
	
	public void marshal(Object source, HierarchicalStreamWriter writer, MarshallingContext context) {
		this.writer=writer;
		this.marshallingContext=context;
		
		SceneComponent src = (SceneComponent) source;
		
		write("icons", src.getIcons());
		writeInterface("shape", src.getShape() );
		write("points", src.getPoints() );
		write("annotations", src.getAnnotations() );
		writeInterface("paint", src.getPaint() );
		writeInterface("outlinePaint", src.getOutlinePaint() );
		writeInterface("pointPaint", src.getPointPaint() );
		writeInterface("annotationPaint", src.getAnnotationPaint() );
		writeInterface("stroke", src.getStroke() );
		writeInterface("pointStroke", src.getPointStroke() );
		writeInterface("pointShape", src.getPointShape() );
		write("annotationMargin", src.getAnnotationMargin() );
		write("font", src.getFont() );
		write("filled", src.getFilled() );
		write("outlined", src.getOutlined() );
		write("pointFilled", src.getPointFilled() );
		write("pointOutlined", src.getPointOutlined() );
		write("annotated", src.getAnnotated() );
		write("areaDragEnabled", src.isAreaDragEnabled() );
		write("outlineDragEnabled", src.isOutlineDragEnabled() );
		write("pointDragEnabled", src.isPointDragEnabled() );
		write("mouseOverEnabled", src.isMouseOverEnabled() );
		write("visible", src.isVisible() );
		write("name", src.getName() );
		write("transform", src.getTransform() );
		
		writer.startNode("children");
		for (int i = 0; i < src.getChildCount(); i++) {
			write("child", src.getChild(i));
		}
		writer.endNode();
		
	}
	
	public Object unmarshal(HierarchicalStreamReader reader, UnmarshallingContext context) {
		this.reader = reader;
		this.context = context;
		
		SceneComponent src = new SceneComponent();
		src.getIcons().clear();
		src.getIcons().addAll( (List) read() );
		src.setShape((Shape) read());
		src.getPoints().clear();
		src.getPoints().addAll((List) read());
		src.getAnnotations().clear();
		src.getAnnotations().addAll((List) read());
		src.setPaint((Paint) read());
		src.setOutlinePaint((Paint) read());
		src.setPointPaint((Paint) read());
		src.setAnnotationPaint((Paint) read());
		src.setStroke((Stroke) read());
		src.setPointStroke((Stroke) read());
		src.setPointShape((Shape) read());
		src.setAnnotationMargin((Integer) read());
		src.setFont((Font) read());
		src.setFilled((Boolean) read());
		src.setOutlined((Boolean) read());
		src.setPointFilled((Boolean) read());
		src.setPointOutlined((Boolean) read());
		src.setAnnotated((Boolean) read());
		src.setAreaDragEnabled(readBool());
		src.setOutlineDragEnabled(readBool());
		src.setPointDragEnabled(readBool());
		src.setMouseOverEnabled(readBool());
		src.setVisible(readBool());
		src.setName((String) read());
		src.setTransform((AffineTransform) read());
		
		reader.moveDown();
	    while (reader.hasMoreChildren()) {
	      src.addChild((SceneComponent) read());
	    }
	    
	    reader.moveUp();
		return src;
	}
	
	private void write(String name, double d) {
		write(name, new java.lang.Double(d));
	}
	
	private void writeInterface(String name, Object src) {
		writer.startNode(name);
		if (src != null) {
			writer.addAttribute("class", mapper.serializedClass(src.getClass()));
			marshallingContext.convertAnother(src);
		}
		writer.endNode();
	}
	
	private void write(String name, int i) {
		write(name, new Integer(i));
	}
	
	private void write(String name, boolean b) {
		write(name, Boolean.valueOf(b));
	}
	
	private void write(String name, Object src) {
		writer.startNode(name);
		if (src != null) marshallingContext.convertAnother(src);
		writer.endNode();
	}
	
	private boolean readBool() {
		return ((Boolean)read(Boolean.class)).booleanValue();
	}

	private Object read() {
 		Object ret = null;
		Class type = null;
		reader.moveDown();
		String className = reader.getAttribute("class");
		if (className != null)
			type = mapper.realClass(className);
		try {
			ret = context.convertAnother(null, type);
			reader.moveUp();
		} catch (ConversionException e) {
			if (e.getCause() instanceof NoSuchElementException) {
				reader.moveUp();
				return null;
			}
			else throw e;
		}
		return ret;
	}
	private Object read(Class type) {
		Object ret = null;
		reader.moveDown();
		try {
			ret = context.convertAnother(null, type);
			reader.moveUp();
		} catch (ConversionException e) {
			if (e.getCause() instanceof NoSuchElementException) {
				reader.moveUp();
				return null;
			}
			else throw e;
		}
		return ret;
	}
}
