/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.modelling;


import de.jtem.java2dx.Point2DList;
import de.jtem.java2dx.Polygon2D;
import de.jtem.java2dx.numerical.PolygonSubdivider;

public class SubdividedPolygon2D extends DraggablePolygon2D {

	private static final long serialVersionUID = -8133229258612452442L;
	protected PolygonSubdivider subdivider;
	private int storedSubdivisionMethod;
	private double storedMaximalCurvatureLength;
	private double storedMaximalEdgeLength;
	
	public SubdividedPolygon2D() {
		this(3,DraggablePoint2DList.CIRCULAR,0,0,1,true,true,.1,.4,PolygonSubdivider.CIRCULAR);
	}
	
	public SubdividedPolygon2D(
			int defaultPointCount,
			int defaultArrangement,
			double defaultTranslationX,
			double defaultTranslationY,
			double defaultRadius,
			boolean closed,
			boolean drawingPoints,
			double maximalCurvatureLength,
			double maximalEdgeLength,
			int subdivisionMethod
	) {
		super(
				defaultPointCount,
				defaultArrangement,
				defaultTranslationX,
				defaultTranslationY,
				defaultRadius,
				closed,
				drawingPoints
		);
		subdivider = new PolygonSubdivider();
		subdivider.setMaximalCurvatureLength(maximalCurvatureLength);
		subdivider.setMaximalEdgeLength(maximalEdgeLength);
		subdivider.setSubdivisionMethod(subdivisionMethod);
		subdivider.setClosed(closed);
		storedMaximalCurvatureLength = maximalCurvatureLength;
		storedMaximalEdgeLength = maximalEdgeLength;
		storedSubdivisionMethod = subdivisionMethod;
		updateModelFromController();
		updateViewFromModel();
	}
	
	public SubdividedPolygon2D(SubdividedPolygon2D p) {
		super(p);
		subdivider = new PolygonSubdivider();
		subdivider.setMaximalCurvatureLength(p.getMaximalCurvatureLength());
		subdivider.setMaximalEdgeLength(p.getMaximalEdgeLength());
		subdivider.setSubdivisionMethod(p.getSubdivisionMethod());
		subdivider.setClosed(((Polygon2D)p.getModel()).isClosed());
		storedMaximalCurvatureLength = p.getMaximalCurvatureLength();
		storedMaximalEdgeLength = p.getMaximalEdgeLength();
		storedSubdivisionMethod = p.getSubdivisionMethod();
		updateModelFromController();
		updateViewFromModel();
	}

	public double getMaximalCurvatureLength() {
		return subdivider.getMaximalCurvatureLength();
	}

	public double getMaximalEdgeLength() {
		return subdivider.getMaximalEdgeLength();
	}

	public int getSubdivisionMethod() {
		return subdivider.getSubdivisionMethod();
	}

	public void setMaximalCurvatureLength(double l) {
		subdivider.setMaximalCurvatureLength(l);
		updateModelFromController();
		fireModelChange();
		updateViewFromModel();
	}

	public void setMaximalEdgeLength(double l) {
		subdivider.setMaximalEdgeLength(l);
		updateModelFromController();
		fireModelChange();
		updateViewFromModel();
	}

	public void setSubdivisionMethod(int subdivisionMethod) {
		subdivider.setSubdivisionMethod(subdivisionMethod);
		updateModelFromController();
		fireModelChange();
		updateViewFromModel();
	}
	
	protected void updateModelFromController() {
		Point2DList p = getControlPoints().getModel();
		Polygon2D polygon = getModel();
		boolean closed = polygon.isClosed();
		subdivider.setClosed(closed);
		polygon.assign(subdivider.subdivide(p.toDoubleArray()),closed);
	}
	
	protected void updateFromModel() {
		// polygon might have changed its closed property
		// updateModelFromController();
		Polygon2D p = getModel();
		p.assign(subdivider.subdivide(p.toDoubleArray()),p.isClosed());
		fireModelChange();
		updateViewFromModel();
	}
	
	public void storeModelState() {
		super.storeModelState();
		storedSubdivisionMethod = subdivider.getSubdivisionMethod();
		storedMaximalCurvatureLength = subdivider.getMaximalCurvatureLength();
		storedMaximalEdgeLength = subdivider.getMaximalEdgeLength();
	}
	
	public void recallModelState() {
		subdivider.setSubdivisionMethod(storedSubdivisionMethod);
		subdivider.setMaximalCurvatureLength(storedMaximalCurvatureLength);
		subdivider.setMaximalEdgeLength(storedMaximalEdgeLength);
		super.recallModelState();
	}
	
	public SubdividedPolygon2D clone() {
		return new SubdividedPolygon2D(this);
	}
}
